#include "Cube.hpp"
#include <glm\gtc\matrix_transform.hpp>


using namespace scene;

Cube::Cube(glm::mat4& matrix, types::Shader* _shader, types::Texture* _texture) :Mesh(matrix),
shader(_shader),
texture(_texture)
{
	//generate vertex buffers and upload data, once for each attribute
	glGenBuffers(1, &positionBuffer);
	glBindBuffer(GL_ARRAY_BUFFER, positionBuffer);
	glBufferData(GL_ARRAY_BUFFER, CUBE_VERTEX_COUNT * 3 * sizeof(float), positions, GL_STATIC_DRAW);
	glBindBuffer(GL_ARRAY_BUFFER, 0);

	glGenBuffers(1, &indexBuffer);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, indexBuffer);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, CUBE_INDEX_COUNT * sizeof(unsigned int), indices, GL_STATIC_DRAW);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

	glGenBuffers(1, &normalsBuffer);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, normalsBuffer);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, CUBE_VERTEX_COUNT * 3 * sizeof(float), normals, GL_STATIC_DRAW);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

	glGenBuffers(1, &uvBuffer);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, uvBuffer);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, CUBE_VERTEX_COUNT * 2 * sizeof(float), uvs, GL_STATIC_DRAW);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);


	//generate bindings
	glGenVertexArrays(1, &vao);
	glBindVertexArray(vao);

	glBindBuffer(GL_ARRAY_BUFFER, positionBuffer);
	GLint positionIndex = glGetAttribLocation(shader->programHandle, "position");
	glEnableVertexAttribArray(positionIndex);
	glVertexAttribPointer(positionIndex, 3, GL_FLOAT, GL_FALSE, 0, 0);

	glBindBuffer(GL_ARRAY_BUFFER, normalsBuffer);
	GLint normalIndex = glGetAttribLocation(shader->programHandle, "normal");
	glEnableVertexAttribArray(normalIndex);
	glVertexAttribPointer(normalIndex, 3, GL_FLOAT, GL_FALSE, 0, 0);

	glBindBuffer(GL_ARRAY_BUFFER, uvBuffer);
	GLint uvlIndex = glGetAttribLocation(shader->programHandle, "uv");
	glEnableVertexAttribArray(uvlIndex);
	glVertexAttribPointer(uvlIndex, 2, GL_FLOAT, GL_FALSE, 0, 0);

	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, indexBuffer);

	glBindVertexArray(0);
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

}

Cube::~Cube(){
	glDeleteBuffers(1, &uvBuffer);
	glDeleteBuffers(1, &positionBuffer);
	glDeleteBuffers(1, &normalsBuffer);
	glDeleteBuffers(1, &indexBuffer);
	glDeleteBuffers(1, &vao);
}
void Cube::update(float deltaT){
	modelMatrix = glm::rotate(modelMatrix, deltaT, glm::vec3(0, 1, 0));
}
void Cube::draw(Shadow* shadow){
	shader->useShader(modelMatrix);
	if (texture != nullptr){
		texture->bind(0);
		auto tex_location = glGetUniformLocation(shader->programHandle, "color_texture");
		glUniform1i(tex_location, 0);
	}
	glm::mat4 lightProjection, lightView;
	glm::mat4 lightSpaceMatrix;
	glm::vec3 lightInvDir = glm::vec3(1, 0.5, 0);
	glm::vec3 lightPos = glm::vec3(6, 2, 2);

	glm::mat4 biasmatrix(
		0.5, 0.0, 0.0, 0.0,
		0.0, 0.5, 0.0, 0.0,
		0.0, 0.0, 0.5, 0.0,
		0.5, 0.5, 0.5, 1.0
		);
	GLfloat near_plane = 0.1f, far_plane = 50.0f;

	lightProjection = glm::perspective(glm::radians(60.0f), (16.0f / 9.0f), near_plane, far_plane);
	//glm::ortho(-10.0f, 10.0f, -10.0f, 10.0f, near_plane, far_plane); // glm::perspective(glm::radians(60.0f), (16.0f / 9.0f), near_plane, far_plane);// glm::ortho(-10.0f, 10.0f, -10.0f, 10.0f, near_plane, far_plane); //camera->projection();// glm::perspective(glm::radians(60.0f), (16.0f / 9.0f), near_plane, far_plane);// 
	lightView = glm::lookAt(lightPos, lightPos - lightInvDir, glm::vec3(0, 1, 0));
	lightSpaceMatrix = lightProjection * lightView;
	lightSpaceMatrix = biasmatrix*lightSpaceMatrix;
	GLuint depthMatrixID = glGetUniformLocation(shader->programHandle, "depthMVP");
	glUniformMatrix4fv(depthMatrixID, 1, GL_FALSE, &lightSpaceMatrix[0][0]);

	shadow->bindForReading(GL_TEXTURE2);
	glUniform1i(glGetUniformLocation(shader->programHandle, "shadowMap"), 2);

	glBindVertexArray(vao);
	glDrawElements(GL_TRIANGLES, CUBE_INDEX_COUNT, GL_UNSIGNED_INT, 0);
	glBindVertexArray(0);
}

void Cube::drawShadow(Shader* shadowShader, Shadow* shadow){
	shadowShader->useShader(modelMatrix);
	//glEnable(GL_POLYGON_OFFSET_FILL);
	//glPolygonOffset(30.0f, 0.0f);
	glm::mat4 lightProjection, lightView;
	glm::mat4 lightSpaceMatrix;
	glm::vec3 lightInvDir = glm::vec3(1, 0.5, 0);
	glm::vec3 lightPos = glm::vec3(6, 2, 2);
	GLfloat near_plane = 0.1f, far_plane = 50.0f;
	lightProjection = glm::perspective(glm::radians(60.0f), (16.0f / 9.0f), near_plane, far_plane);
	// glm::ortho(-10.0f, 10.0f, -10.0f, 10.0f, near_plane, far_plane); //glm::perspective(glm::radians(60.0f), (16.0f / 9.0f), near_plane, far_plane);//glm::perspective(glm::radians(60.0f), (float)(800 / 600), near_plane, far_plane);//  //camera->projection();
	lightView = glm::lookAt(lightPos, lightPos - lightInvDir, glm::vec3(0, 1, 0));
	lightSpaceMatrix = lightProjection * lightView;//*modelMatrix;


	GLuint depthMatrixID = glGetUniformLocation(shadowShader->programHandle, "depthMVP");
	glUniformMatrix4fv(depthMatrixID, 1, GL_FALSE, &lightSpaceMatrix[0][0]);

	glBindVertexArray(vao);
	glBindBuffer(GL_ARRAY_BUFFER, positionBuffer);
	GLint positionIndex = glGetAttribLocation(shadowShader->programHandle, "position");
	glEnableVertexAttribArray(positionIndex);
	glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, (void*)0);
	glDrawElements(GL_TRIANGLES, CUBE_INDEX_COUNT, GL_UNSIGNED_INT, 0);
	glBindVertexArray(0);
}

const float Cube::positions[CUBE_VERTEX_COUNT * 3] = {

	//Back
	-0.5f, -0.5f, -0.5f,
	-0.5f, 0.5f, -0.5f,
	0.5f, 0.5f, -0.5f,
	0.5f, -0.5f, -0.5f,

	//Front
	-0.5f, -0.5f, 0.5f,
	0.5f, -0.5f, 0.5f,
	0.5f, 0.5f, 0.5f,
	-0.5f, 0.5f, 0.5f,

	//Top
	0.5f, 0.5f, 0.5f,
	0.5f, 0.5f, -0.5f,
	-0.5f, 0.5f, -0.5f,
	-0.5f, 0.5f, 0.5f,

	//Bottom
	0.5f, -0.5f, 0.5f,
	-0.5f, -0.5f, 0.5f,
	-0.5f, -0.5f, -0.5f,
	0.5f, -0.5f, -0.5f,

	//Right
	0.5f, 0.5f, 0.5f,
	0.5f, -0.5f, 0.5f,
	0.5f, -0.5f, -0.5f,
	0.5f, 0.5f, -0.5f,

	//Left
	-0.5f, 0.5f, 0.5f,
	-0.5f, 0.5f, -0.5f,
	-0.5f, -0.5f, -0.5f,
	-0.5f, -0.5f, 0.5f,


};

const float Cube::normals[CUBE_VERTEX_COUNT * 3] = {
	0.0f, 0.0f, -1.0f,
	0.0f, 0.0f, -1.0f,
	0.0f, 0.0f, -1.0f,
	0.0f, 0.0f, -1.0f,

	0.0f, 0.0f, 1.0f,
	0.0f, 0.0f, 1.0f,
	0.0f, 0.0f, 1.0f,
	0.0f, 0.0f, 1.0f,

	0.0f, 1.0f, 0.0f,
	0.0f, 1.0f, 0.0f,
	0.0f, 1.0f, 0.0f,
	0.0f, 1.0f, 0.0f,

	0.0f, -1.0f, 0.0f,
	0.0f, -1.0f, 0.0f,
	0.0f, -1.0f, 0.0f,
	0.0f, -1.0f, 0.0f,

	1.0f, 0.0f, 0.0f,
	1.0f, 0.0f, 0.0f,
	1.0f, 0.0f, 0.0f,
	1.0f, 0.0f, 0.0f,

	-1.0f, 0.0f, 0.0f,
	-1.0f, 0.0f, 0.0f,
	-1.0f, 0.0f, 0.0f,
	-1.0f, 0.0f, 0.0f,
};

const unsigned int  Cube::indices[CUBE_INDEX_COUNT] = {
	0, 1, 2,
	0, 2, 3,

	4, 5, 6,
	4, 6, 7,

	8, 9, 10,
	8, 10, 11,

	12, 13, 14,
	12, 14, 15,

	16, 17, 18,
	16, 18, 19,

	20, 21, 22,
	20, 22, 23
};


const float Cube::uvs[CUBE_VERTEX_COUNT * 2] = {
	1.0f, 0.0f,
	1.0f, 1.0f,
	0.0f, 1.0f,
	0.0f, 0.0f,
	
	0.0f, 0.0f,
	1.0f, 0.0f,
	1.0f, 1.0f,
	0.0f, 1.0f,

	0.0f, 1.0f,
	0.0f, 0.0f,
	1.0f, 0.0f,
	1.0f, 1.0f,

	1.0f, 1.0f,
	0.0f, 1.0f,
	0.0f, 0.0f,
	1.0f, 0.0f,

	1.0f, 0.0f,
	1.0f, 1.0f,
	0.0f, 1.0f,
	0.0f, 0.0f,

	0.0f, 0.0f,
	1.0f, 0.0f,
	1.0f, 1.0f,
	0.0f, 1.0f,
};